package gregapi.tileentity.machines;

import static gregapi.data.CS.*;
import gregapi.GT_API;
import gregapi.block.multitileentity.MultiTileEntityRegistry;
import gregapi.code.TagData;
import gregapi.data.CS.ModIDs;
import gregapi.data.CS.SFX;
import gregapi.data.*;
import gregapi.data.LH.Chat;
import gregapi.gui.*;
import gregapi.old.Textures;
import gregapi.recipes.Recipe;
import gregapi.recipes.Recipe.RecipeMap;
import gregapi.render.BlockTextureDefault;
import gregapi.render.BlockTextureMulti;
import gregapi.render.IIconContainer;
import gregapi.render.ITexture;
import gregapi.tileentity.ITileEntityAdjacentInventoryUpdatable;
import gregapi.tileentity.base.TileEntityBase09FacingSingle;
import gregapi.tileentity.data.ITileEntityProgress;
import gregapi.tileentity.delegate.DelegatorTileEntity;
import gregapi.tileentity.energy.ITileEntityEnergy;
import gregapi.util.UT;

import java.util.Collection;
import java.util.List;

import net.minecraft.block.Block;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.InventoryPlayer;
import net.minecraft.inventory.ICrafting;
import net.minecraft.inventory.IInventory;
import net.minecraft.inventory.ISidedInventory;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraftforge.fluids.*;
import buildcraft.api.tiles.IHasWork;
import cpw.mods.fml.common.Optional;
import cpw.mods.fml.relauncher.Side;
import cpw.mods.fml.relauncher.SideOnly;

/**
 * @author Gregorius Techneticies
 * 
 * This is the Base Class for almost all my Basic Machines. It is almost too simple to use.
 * 
 * In order to create a Basic Machine (Steel Shredder in this example), I use the following NBT Parameters in my MultiTileEntity System:
 * 
 * NBT_MATERIAL						= MT.Steel
 * NBT_HARDNESS						= 6.0F
 * NBT_RESISTANCE					= 6.0F
 * NBT_COLOR						= UT.Code.getRGBInt(MT.Steel.mRGBaSolid)
 * NBT_INPUT						= 128
 * NBT_GUI							= RES_PATH_GUI+"machines/Shredder"
 * NBT_TEXTURE						= "shredder"
 * NBT_ENERGY_ACCEPTED				= TD.Energy.RU
 * NBT_RECIPEMAP					= RecipeMap.sShredderRecipes
 * NBT_EFFICIENCY					= 10000
 * NBT_INV_SIDE_INPUT				= SIDE_BITS[SIDE_TOP]
 * NBT_INV_SIDE_AUTO_INPUT			= SIDE_TOP
 * NBT_INV_SIDE_OUTPUT				= SIDE_BITS[SIDE_BOTTOM]
 * NBT_INV_SIDE_AUTO_OUTPUT			= SIDE_BOTTOM
 * NBT_ENERGY_ACCEPTED_SIDES		= SIDE_BITS[SIDE_LEFT]|SIDE_BITS[SIDE_RIGHT]
 */
@Optional.InterfaceList(value = {
	@Optional.Interface(iface = "buildcraft.api.tiles.IHasWork", modid = ModIDs.BC)
})
public class MultiTileEntityBasicMachine extends TileEntityBase09FacingSingle implements IHasWork, ITileEntitySwitchableOnOff, ITileEntityRunningActively, ITileEntityAdjacentInventoryUpdatable, ITileEntityEnergy, ITileEntityProgress, ISidedInventory, IFluidHandler {
	protected boolean mCouldUseRecipe = F, mStopped = F, oActive = F, oRunning = F, mStateNew = F, mStateOld = F, mDisabledItemInput = F, mDisabledItemOutput = F, mDisabledFluidInput = F, mDisabledFluidOutput = F;
	protected byte mEnergyInputs = (byte)(SIDE_BITS[SIDE_BACK] | SIDE_BITS[SIDE_ANY]), mOutputBlocked = 0, mMode = 0;
	protected byte mItemInputs  = (byte)(SIDE_BITS[SIDE_TOP] | SIDE_BITS[SIDE_ANY]), mItemOutputs  = (byte)(SIDE_BITS[SIDE_BOTTOM] | SIDE_BITS[SIDE_ANY]), mItemAutoInput  = SIDE_TOP, mItemAutoOutput  = SIDE_BOTTOM;
	protected byte mFluidInputs = (byte)(SIDE_BITS[SIDE_TOP] | SIDE_BITS[SIDE_ANY]), mFluidOutputs = (byte)(SIDE_BITS[SIDE_BOTTOM] | SIDE_BITS[SIDE_ANY]), mFluidAutoInput = SIDE_TOP, mFluidAutoOutput = SIDE_BOTTOM;
	protected short mEfficiency = 10000;
	protected long mEnergy = 0, mInputMin = 16, mInput = 32, mInputMax = 64, mMinEnergy = 0;
	protected TagData mEnergyTypeAccepted = TD.Energy.TU;
	protected Recipe mLastRecipe = null;
	protected FluidTank[] mTanksInput = ZL_FLUIDTANK, mTanksOutput = ZL_FLUIDTANK;
	protected ItemStack[] mOutputItems = ZL_ITEMSTACK;
	protected FluidStack[] mOutputFluids = ZL_FLUIDSTACK;
	protected IIconContainer[] mTexturesMaterial = L6_IICONCONTAINER, mTexturesInactive = L6_IICONCONTAINER, mTexturesActive = L6_IICONCONTAINER, mTexturesRunning = L6_IICONCONTAINER;
	
	public String mGUITexture = "";
	public RecipeMap mRecipes = RecipeMap.sFurnaceRecipes;
	public long mProgress = 0, mMaxProgress = 0;
	public boolean mSuccessful = F, mActive = F, mRunning = F;
	
    @Override
	public void readFromNBT2(NBTTagCompound aNBT) {
    	super.readFromNBT2(aNBT);
    	mEnergy = aNBT.getLong(NBT_ENERGY);
        if (aNBT.hasKey(NBT_ACTIVE)) mCouldUseRecipe = mActive = aNBT.getBoolean(NBT_ACTIVE);
        if (aNBT.hasKey(NBT_STOPPED)) mStopped = aNBT.getBoolean(NBT_STOPPED);
        if (aNBT.hasKey(NBT_RUNNING)) mRunning = aNBT.getBoolean(NBT_RUNNING);
        if (aNBT.hasKey(NBT_STATE+".new")) mStateNew = aNBT.getBoolean(NBT_STATE+".new");
        if (aNBT.hasKey(NBT_STATE+".old")) mStateOld = aNBT.getBoolean(NBT_STATE+".old");
        if (aNBT.hasKey(NBT_EFFICIENCY)) mEfficiency = (short)UT.Code.bind(0, 10000, aNBT.getShort(NBT_EFFICIENCY));
        if (aNBT.hasKey(NBT_INPUT)) {mInput = aNBT.getLong(NBT_INPUT); mInputMin = mInput / 2; mInputMax = mInput * 2;}
        if (aNBT.hasKey(NBT_INPUT_MIN)) {mInputMin = aNBT.getLong(NBT_INPUT_MIN);}
        if (aNBT.hasKey(NBT_INPUT_MAX)) {mInputMax = aNBT.getLong(NBT_INPUT_MAX);}
        if (aNBT.hasKey(NBT_MINENERGY)) {mMinEnergy = aNBT.getLong(NBT_MINENERGY);}
        if (aNBT.hasKey(NBT_PROGRESS)) {mProgress = aNBT.getLong(NBT_PROGRESS);}
        if (aNBT.hasKey(NBT_MAXPROGRESS)) {mMaxProgress = aNBT.getLong(NBT_MAXPROGRESS);}
        if (aNBT.hasKey(NBT_MODE)) mMode = aNBT.getByte(NBT_MODE);
        if (aNBT.hasKey(NBT_INV_SIDE_INPUT)) mItemInputs = (byte)(aNBT.getByte(NBT_INV_SIDE_INPUT) | SIDE_BITS[SIDE_ANY]);
        if (aNBT.hasKey(NBT_INV_SIDE_OUTPUT)) mItemOutputs = (byte)(aNBT.getByte(NBT_INV_SIDE_OUTPUT) | SIDE_BITS[SIDE_ANY]);
        if (aNBT.hasKey(NBT_INV_SIDE_AUTO_INPUT)) mItemAutoInput = aNBT.getByte(NBT_INV_SIDE_AUTO_INPUT);
        if (aNBT.hasKey(NBT_INV_SIDE_AUTO_OUTPUT)) mItemAutoOutput = aNBT.getByte(NBT_INV_SIDE_AUTO_OUTPUT);
        if (aNBT.hasKey(NBT_INV_DISABLED_INPUT)) mDisabledItemInput = aNBT.getBoolean(NBT_INV_DISABLED_INPUT);
        if (aNBT.hasKey(NBT_INV_DISABLED_OUTPUT)) mDisabledItemOutput = aNBT.getBoolean(NBT_INV_DISABLED_OUTPUT);
        if (aNBT.hasKey(NBT_TANK_SIDE_INPUT)) mFluidInputs = (byte)(aNBT.getByte(NBT_TANK_SIDE_INPUT) | SIDE_BITS[SIDE_ANY]);
        if (aNBT.hasKey(NBT_TANK_SIDE_OUTPUT)) mFluidOutputs = (byte)(aNBT.getByte(NBT_TANK_SIDE_OUTPUT) | SIDE_BITS[SIDE_ANY]);
        if (aNBT.hasKey(NBT_TANK_SIDE_AUTO_INPUT)) mFluidAutoInput = aNBT.getByte(NBT_TANK_SIDE_AUTO_INPUT);
        if (aNBT.hasKey(NBT_TANK_SIDE_AUTO_OUTPUT)) mFluidAutoOutput = aNBT.getByte(NBT_TANK_SIDE_AUTO_OUTPUT);
        if (aNBT.hasKey(NBT_TANK_DISABLED_INPUT)) mDisabledFluidInput = aNBT.getBoolean(NBT_TANK_DISABLED_INPUT);
        if (aNBT.hasKey(NBT_TANK_DISABLED_OUTPUT)) mDisabledFluidOutput = aNBT.getBoolean(NBT_TANK_DISABLED_OUTPUT);
        if (aNBT.hasKey(NBT_ENERGY_ACCEPTED)) mEnergyTypeAccepted = TagData.createTagData(aNBT.getString(NBT_ENERGY_ACCEPTED));
        if (aNBT.hasKey(NBT_ENERGY_ACCEPTED_SIDES)) mEnergyInputs = (byte)(aNBT.getByte(NBT_ENERGY_ACCEPTED_SIDES) | SIDE_BITS[SIDE_ANY]);
        
        int tCapacity = Math.max(1000, mRecipes.mMaxFluidInputSize * 2);
        if (aNBT.hasKey(NBT_TANK_CAPACITY)) tCapacity = UT.Code.bindInt(aNBT.getLong(NBT_TANK_CAPACITY));
        mTanksInput = new FluidTank[mRecipes.mInputFluidCount];
        for (int i = 0; i < mTanksInput.length; i++) mTanksInput[i] = new FluidTank(tCapacity).readFromNBT(aNBT.getCompoundTag(NBT_TANK+".in."+i));
        mTanksOutput = new FluidTank[mRecipes.mOutputFluidCount];
        for (int i = 0; i < mTanksOutput.length; i++) mTanksOutput[i] = new FluidTank(tCapacity).readFromNBT(aNBT.getCompoundTag(NBT_TANK+".out."+i));
        
        mOutputFluids = new FluidStack[mRecipes.mOutputFluidCount];
		for (int i = 0; i < mOutputFluids.length; i++) mOutputFluids[i] = FluidStack.loadFluidStackFromNBT(aNBT.getCompoundTag(NBT_TANK_OUTPUT+"."+i));
		
        mOutputItems = new ItemStack[mRecipes.mOutputItemsCount];
		for (int i = 0; i < mOutputItems.length; i++) mOutputItems[i] = UT.Stacks.load(aNBT, NBT_INV_OUTPUT+"."+i);
		
		if (CODE_CLIENT) {
	        if (aNBT.hasKey(NBT_GUI)) {
	        	mGUITexture = aNBT.getString(NBT_GUI);
		        if (!mGUITexture.endsWith(".png")) mGUITexture += ".png";
	        }
			if (GT_API.sBlockIcons == null && aNBT.hasKey(NBT_TEXTURE)) {
				String tTextureName = aNBT.getString(NBT_TEXTURE);
				mTexturesMaterial = new IIconContainer[] {
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/bottom"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/top"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/left"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/front"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/right"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/back")};
				mTexturesInactive = new IIconContainer[] {
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/bottom"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/top"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/left"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/front"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/right"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/back")};
				mTexturesActive = new IIconContainer[] {
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/bottom"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/top"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/left"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/front"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/right"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/back")};
				mTexturesRunning = new IIconContainer[] {
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/bottom"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/top"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/left"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/front"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/right"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/back")};
			} else {
		        if (getMultiTileEntityRegistryID() != W && getMultiTileEntityID() != W) {
					MultiTileEntityBasicMachine tCanonicalTileEntity = (MultiTileEntityBasicMachine)MultiTileEntityRegistry.getRegistry(getMultiTileEntityRegistryID()).getClassContainer(getMultiTileEntityID()).mCanonicalTileEntity;
					mTexturesMaterial = tCanonicalTileEntity.mTexturesMaterial;
					mTexturesInactive = tCanonicalTileEntity.mTexturesInactive;
					mTexturesRunning  = tCanonicalTileEntity.mTexturesRunning;
					mTexturesActive   = tCanonicalTileEntity.mTexturesActive;
		        } else {
		        	mTexturesMaterial = mTexturesInactive = mTexturesRunning = mTexturesActive = new IIconContainer[6];
		        }
	        }
		}
	}
    
    @Override
	public void writeToNBT2(NBTTagCompound aNBT) {
    	super.writeToNBT2(aNBT);
    	UT.NBT.setNumber(aNBT, NBT_ENERGY, mEnergy);
    	UT.NBT.setNumber(aNBT, NBT_MINENERGY, mMinEnergy);
    	UT.NBT.setNumber(aNBT, NBT_PROGRESS, mProgress);
    	UT.NBT.setNumber(aNBT, NBT_MAXPROGRESS, mMaxProgress);
    	UT.NBT.setBoolean(aNBT, NBT_ACTIVE, mActive);
    	UT.NBT.setBoolean(aNBT, NBT_RUNNING, mRunning);
    	UT.NBT.setBoolean(aNBT, NBT_STOPPED, mStopped);
    	UT.NBT.setBoolean(aNBT, NBT_STATE+".new", mStateNew);
    	UT.NBT.setBoolean(aNBT, NBT_STATE+".old", mStateOld);
    	
    	UT.NBT.setNumber(aNBT, NBT_MODE, mMode);
    	UT.NBT.setBoolean(aNBT, NBT_INV_DISABLED_INPUT, mDisabledItemInput);
    	UT.NBT.setBoolean(aNBT, NBT_INV_DISABLED_OUTPUT, mDisabledItemOutput);
    	UT.NBT.setBoolean(aNBT, NBT_TANK_DISABLED_INPUT, mDisabledFluidInput);
    	UT.NBT.setBoolean(aNBT, NBT_TANK_DISABLED_OUTPUT, mDisabledFluidOutput);
    	
    	for (int i = 0; i < mTanksInput .length; i++) if (mTanksInput [i] != null) aNBT.setTag(NBT_TANK+".in." +i, mTanksInput [i].writeToNBT(new NBTTagCompound()));
    	for (int i = 0; i < mTanksOutput.length; i++) if (mTanksOutput[i] != null) aNBT.setTag(NBT_TANK+".out."+i, mTanksOutput[i].writeToNBT(new NBTTagCompound()));
    	for (int i = 0; i < mOutputFluids.length; i++) if (mOutputFluids[i] != null) aNBT.setTag(NBT_TANK_OUTPUT+"."+i, mOutputFluids[i].writeToNBT(new NBTTagCompound()));
    	for (int i = 0; i < mOutputItems.length; i++) UT.Stacks.save(aNBT, NBT_INV_OUTPUT+"."+i, mOutputItems[i]);
    }
    
	@Override
    public NBTTagCompound writeItemNBT2(NBTTagCompound aNBT) {
    	super.writeItemNBT2(aNBT);
    	UT.NBT.setNumber(aNBT, NBT_MODE, mMode);
    	UT.NBT.setBoolean(aNBT, NBT_INV_DISABLED_INPUT, mDisabledItemInput);
    	UT.NBT.setBoolean(aNBT, NBT_INV_DISABLED_OUTPUT, mDisabledItemOutput);
    	UT.NBT.setBoolean(aNBT, NBT_TANK_DISABLED_INPUT, mDisabledFluidInput);
    	UT.NBT.setBoolean(aNBT, NBT_TANK_DISABLED_OUTPUT, mDisabledFluidOutput);
		return aNBT;
	}
	
	@Override
	public void addToolTips(List aList, ItemStack aStack, boolean aF3_H) {
		aList.add(Chat.CYAN + LH.get(LH.RECIPES) + ": " + Chat.WHITE + LH.get(mRecipes.mNameInternal));
		aList.add(LH.getToolTipEfficiency(mEfficiency));
		
		String tSideNames = ""; boolean temp = F;
		if (mEnergyTypeAccepted != TD.Energy.TU) {
		for (byte tSide : ALL_SIDES_VALID) if (FACE_CONNECTED[tSide][mEnergyInputs])	{tSideNames += (temp?", ":"")+LH.get(LH.FACES[tSide]); temp = T;}
		LH.addEnergyToolTips(this, aList, mEnergyTypeAccepted, null, tSideNames, null);
		}
		
		tSideNames = ""; temp = F;
		for (byte tSide : ALL_SIDES_VALID) if (FACE_CONNECTED[tSide][mItemInputs])		{tSideNames += (temp?", ":"")+LH.get(LH.FACES[tSide]); temp = T;}
		if (UT.Code.stringValid(tSideNames)) aList.add(Chat.GREEN	+ LH.get(LH.ITEM_INPUT)		+ ": " + Chat.WHITE + tSideNames);
		tSideNames = ""; temp = F;
		for (byte tSide : ALL_SIDES_VALID) if (FACE_CONNECTED[tSide][mItemOutputs])		{tSideNames += (temp?", ":"")+LH.get(LH.FACES[tSide]); temp = T;}
		if (UT.Code.stringValid(tSideNames)) aList.add(Chat.RED		+ LH.get(LH.ITEM_OUTPUT)	+ ": " + Chat.WHITE + tSideNames);
		tSideNames = ""; temp = F;
		for (byte tSide : ALL_SIDES_VALID) if (FACE_CONNECTED[tSide][mFluidInputs])		{tSideNames += (temp?", ":"")+LH.get(LH.FACES[tSide]); temp = T;}
		if (UT.Code.stringValid(tSideNames)) aList.add(Chat.GREEN	+ LH.get(LH.FLUID_INPUT)	+ ": " + Chat.WHITE + tSideNames);
		tSideNames = ""; temp = F;
		for (byte tSide : ALL_SIDES_VALID) if (FACE_CONNECTED[tSide][mFluidOutputs])	{tSideNames += (temp?", ":"")+LH.get(LH.FACES[tSide]); temp = T;}
		if (UT.Code.stringValid(tSideNames)) aList.add(Chat.RED		+ LH.get(LH.FLUID_OUTPUT)	+ ": " + Chat.WHITE + tSideNames);
		
		aList.add(Chat.DGRAY	+ LH.get(LH.TOOL_TO_TOGGLE_SCREWDRIVER));
		aList.add(Chat.DGRAY	+ LH.get(LH.TOOL_TO_TOGGLE_AUTO_INPUTS_MONKEY_WRENCH));
		aList.add(Chat.DGRAY	+ LH.get(LH.TOOL_TO_TOGGLE_AUTO_OUTPUTS_MONKEY_WRENCH));
		aList.add(Chat.DGRAY	+ LH.get(LH.TOOL_TO_RESET_SOFT_HAMMER));
		super.addToolTips(aList, aStack, aF3_H);
	}
	
	@Override
	public long onToolClick2(String aTool, long aRemainingDurability, long aQuality, Entity aPlayer, List<String> aChatReturn, IInventory aPlayerInventory, boolean aSneaking, ItemStack aStack, byte aSide, float aHitX, float aHitY, float aHitZ) {
		long rReturn = super.onToolClick2(aTool, aRemainingDurability, aQuality, aPlayer, aChatReturn, aPlayerInventory, aSneaking, aStack, aSide, aHitX, aHitY, aHitZ);
		if (rReturn > 0) return rReturn;
		
		if (isClientSide()) return 0;
		
		if (aTool.equals(TOOL_screwdriver)) {
			mMode = (byte)((mMode + 1) % 2);
			aChatReturn.add((mMode & 1) != 0 ?"Only produce when Output is completely empty":"Produce whenever there is space");
			updateInventory();
			return 10000;
		}
		if (aTool.equals(TOOL_softhammer)) {
			mProgress = mMinEnergy = mMaxProgress = 0;
			mOutputFluids = ZL_FLUIDSTACK;
			mOutputItems = ZL_ITEMSTACK;
			updateInventory();
			return 10000;
		}
		if (aTool.equals(TOOL_monkeywrench)) {
			long rOutput = 0;
			if (FACING_TO_SIDE[mFacing][mItemAutoInput] == aSide) {
				mDisabledItemInput = !mDisabledItemInput;
				aChatReturn.add(mDisabledItemInput?"Auto Item Input Disabled":"Auto Item Input Enabled");
				rOutput += 10000;
			}
			if (FACING_TO_SIDE[mFacing][mItemAutoOutput] == aSide) {
				mDisabledItemOutput = !mDisabledItemOutput;
				aChatReturn.add(mDisabledItemOutput?"Auto Item Output Disabled":"Auto Item Output Enabled");
				rOutput += 10000;
			}
			if (FACING_TO_SIDE[mFacing][mFluidAutoInput] == aSide) {
				mDisabledFluidInput = !mDisabledFluidInput;
				aChatReturn.add(mDisabledFluidInput?"Auto Fluid Input Disabled":"Auto Fluid Input Enabled");
				rOutput += 10000;
			}
			if (FACING_TO_SIDE[mFacing][mFluidAutoOutput] == aSide) {
				mDisabledFluidOutput = !mDisabledFluidOutput;
				aChatReturn.add(mDisabledFluidOutput?"Auto Fluid Output Disabled":"Auto Fluid Output Enabled");
				rOutput += 10000;
			}
			if (rOutput > 0) {
				updateInventory();
				return rOutput;
			}
		}
		return 0;
	}
	
    @Override
	public void onCoordinateChange() {
		updateAdjacentToggleableEnergySources();
	}
	
    @Override
    public void onTickFailed(long aTimer, boolean aIsServerSide) {
    	super.onTickFailed(aTimer, aIsServerSide);
    	// Just to prevent Infinite Item dupes in case this happens during the Processing Functions.
		mProgress = mMinEnergy = mMaxProgress = 0;
		mOutputFluids = ZL_FLUIDSTACK;
		mOutputItems = ZL_ITEMSTACK;
    }
    
	@Override
    public void onTickFirst2(boolean aIsServerSide) {
    	super.onTickFirst2(aIsServerSide);
    	if (aIsServerSide) {
        	updateAdjacentToggleableEnergySources();
        	if (!mActive) checkRecipe(F, mRunning || mStopped);
    	}
    }
	
    @Override
    public void onTick2(long aTimer, boolean aIsServerSide) {
    	if (aIsServerSide) {
    		if (mBlockUpdated) updateAdjacentToggleableEnergySources();
    		if (mEnergyTypeAccepted == TD.Energy.TU && !mStopped) mEnergy++;

			if (!mDisabledFluidOutput && SIDES_VALID[mFluidAutoOutput]) {
				byte tAutoOutput = FACING_TO_SIDE[mFacing][mFluidAutoOutput];
				DelegatorTileEntity<IFluidHandler> tTileEntity = getAdjacentTank(tAutoOutput);
				if (tTileEntity.mTileEntity != null) for (FluidTank tTank : mTanksOutput) {
					FluidStack tDrained = tTank.drain(Integer.MAX_VALUE, F);
					if (tDrained != null) {
						tDrained.amount = tTileEntity.mTileEntity.fill(FORGE_DIR[tTileEntity.mSideOfTileEntity], tDrained, F);
						if (tDrained.amount > 0 && tTileEntity.mTileEntity.fill(FORGE_DIR[tTileEntity.mSideOfTileEntity], tTank.drain(tDrained.amount, T), T) > 0) updateInventory();
					}
				}
			}
			
    		if (mTimer > 2) doWork(aTimer);
    		
			if (mTimer % 600 == 5) doDefaultStructuralChecks();
    		
    		for (int i = 0; i < mTanksInput .length; i++) slot(mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount + 1 + i						, UT.Fluids.display(mTanksInput [i].getFluid(), T));
    		for (int i = 0; i < mTanksOutput.length; i++) slot(mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount + 1 + i + mTanksInput.length	, UT.Fluids.display(mTanksOutput[i].getFluid(), T));
    	}
    }
    
    @Override
    public boolean onTickCheck(long aTimer) {
    	return mActive != oActive || mRunning != oRunning || super.onTickCheck(aTimer);
    }
    
	@Override
    public void onTickResetChecks(long aTimer, boolean aIsServerSide) {
    	super.onTickResetChecks(aTimer, aIsServerSide);
		oRunning = mRunning;
		oActive  = mActive;
    }
    
	@Override
	public boolean onBlockActivated3(EntityPlayer aPlayer, byte aSide, float aHitX, float aHitY, float aHitZ) {
		if (isServerSide()) openGUI(aPlayer, aSide);
		return T;
	}
	
	@Override
	public long doInject(TagData aEnergyType, byte aSide, long aSize, long aAmount, boolean aDoInject) {
		if (mStopped) return 0;
		if (aDoInject) mStateNew = (aSize > 0);
		aSize = Math.abs(aSize);
		if (aSize > getEnergySizeInputMax(aEnergyType, aSide)) {
			if (aDoInject) overcharge(aSize, aEnergyType);
			return aAmount;
		}
		long tInput = Math.min(mInputMax - mEnergy, aSize * aAmount), tConsumed = Math.min(aAmount, (tInput/aSize) + (tInput%aSize!=0?1:0));
		if (aDoInject) mEnergy += tConsumed * aSize;
		return tConsumed;
	}
	
	@Override public boolean isEnergyType					(TagData aEnergyType, byte aSide, boolean aEmitting) {return !aEmitting && aEnergyType == mEnergyTypeAccepted;}
	@Override public boolean isEnergyAcceptingFrom			(TagData aEnergyType, byte aSide, boolean aTheoretical) {return (aTheoretical || !mStopped) && FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mEnergyInputs] && super.isEnergyAcceptingFrom(aEnergyType, aSide, aTheoretical);}
	@Override public long getEnergySizeInputMin				(TagData aEnergyType, byte aSide) {return mInputMin;}
	@Override public long getEnergySizeInputRecommended		(TagData aEnergyType, byte aSide) {return mInput;}
	@Override public long getEnergySizeInputMax				(TagData aEnergyType, byte aSide) {return mInputMax;}
	@Override public Collection<TagData> getEnergyTypes(byte aSide) {return mEnergyTypeAccepted.AS_LIST;}
	
	// Inventory Stuff
	
	@Override
	public ItemStack[] getDefaultInventory(NBTTagCompound aNBT) {
        if (aNBT.hasKey(NBT_RECIPEMAP)) mRecipes = RecipeMap.RECIPE_MAPS.get(aNBT.getString(NBT_RECIPEMAP));
        ACCESSIBLE_SLOTS = new int[mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount];
        for (int i = 0; i < ACCESSIBLE_SLOTS.length; i++) ACCESSIBLE_SLOTS[i] = i;
		return new ItemStack[mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount + 1 + mRecipes.mInputFluidCount + mRecipes.mOutputFluidCount];
	}
	
	protected int[] ACCESSIBLE_SLOTS;
	@Override public int[] getAccessibleSlotsFromSide(int aSide) {return ACCESSIBLE_SLOTS;}
	@Override public boolean canDrop(int aInventorySlot) {return aInventorySlot < mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount + 1;}
	
	@Override
	public boolean canInsertItem(int aSlot, ItemStack aStack, int aSide) {
		if (!FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mItemInputs] || aSlot >= mRecipes.mInputItemsCount) return F;
		for (int i = 0; i < mRecipes.mInputItemsCount; i++) if (UT.Stacks.equal(aStack, slot(i), T)) return i == aSlot;
		return mRecipes.containsInput(aStack);
	}
	
	@Override
	public boolean canExtractItem(int aSlot, ItemStack aStack, int aSide) {
		return FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mItemOutputs] && aSlot >= mRecipes.mInputItemsCount && aSlot < mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount;
	}
	
	// Tank things
	
	@Override
	protected IFluidTank getFluidTankFillable(byte aSide, FluidStack aFluidToFill) {
		if (!FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mFluidInputs]) return null;
		for (int i = 0; i < mTanksInput.length; i++) if (UT.Fluids.equal(aFluidToFill, mTanksInput[i].getFluid(), F)) return mTanksInput[i];
		if (!mRecipes.containsInput(aFluidToFill)) return null;
		for (int i = 0; i < mTanksInput.length; i++) if (mTanksInput[i].getFluidAmount() == 0) return mTanksInput[i];
		return null;
	}
	
	@Override
	protected IFluidTank getFluidTankDrainable(byte aSide, FluidStack aFluidToDrain) {
		if (!FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mFluidOutputs]) return null;
		if (aFluidToDrain == null) {
			for (int i = 0; i < mTanksOutput.length; i++) if (mTanksOutput[i].getFluidAmount() != 0) return mTanksOutput[i];
		} else {
			for (int i = 0; i < mTanksOutput.length; i++) if (UT.Fluids.equal(aFluidToDrain, mTanksOutput[i].getFluid(), F)) return mTanksOutput[i];
		}
		return null;
	}
	
	@Override
	protected IFluidTank[] getFluidTanks(byte aSide) {
		if (FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mFluidInputs]) {
			if (FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mFluidOutputs]) {
				IFluidTank[] rTanks = new IFluidTank[mTanksInput.length + mTanksOutput.length];
				for (int i = 0; i < mTanksInput .length; i++) rTanks[i] = mTanksInput[i];
				for (int i = 0; i < mTanksOutput.length; i++) rTanks[mTanksInput.length+i] = mTanksOutput[i];
				return rTanks;
			}
			return mTanksInput;
		}
		if (FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mFluidOutputs]) return mTanksOutput;
		return ZL_FLUIDTANK;
	}
	
	public void updateAdjacentToggleableEnergySources() {
		for (byte tSide : ALL_SIDES_VALID) if (isEnergyAcceptingFrom(mEnergyTypeAccepted, tSide, T)) {
			DelegatorTileEntity<TileEntity> tDelegator = getAdjacentTileEntity(tSide);
			if (tDelegator.mTileEntity instanceof ITileEntitySwitchableOnOff && tDelegator.mTileEntity instanceof ITileEntityEnergy && ((ITileEntityEnergy)tDelegator.mTileEntity).isEnergyEmittingTo(mEnergyTypeAccepted, tDelegator.mSideOfTileEntity, T)) {
				((ITileEntitySwitchableOnOff)tDelegator.mTileEntity).setStateOnOff(getStateOnOff());
			}
		}
	}
	
	// Stuff to Override
	
	protected boolean canOutput(Recipe aRecipe) {
    	for (int i = 0, j = mRecipes.mInputItemsCount; i < mRecipes.mOutputItemsCount && i < aRecipe.mOutputs.length; i++, j++) if (slot(j) != null) {
        	if ((mMode & 1) != 0 || aRecipe.mNeedsEmptyOutput || (aRecipe.mOutputs[i] != null && (!UT.Stacks.equal(slot(j), aRecipe.mOutputs[i], F) || slot(j).stackSize + aRecipe.mOutputs[i].stackSize > slot(j).getMaxStackSize()))) {
            	mOutputBlocked++;
            	return F;
    		}
    	}
    	for (int i = 0; i < mTanksOutput.length && i < aRecipe.mFluidOutputs.length; i++) if (mTanksOutput[i].getFluidAmount() != 0) {
            if ((mMode & 1) != 0 || aRecipe.mNeedsEmptyOutput || (aRecipe.mFluidOutputs[i] != null && (!UT.Fluids.equal(mTanksOutput[i].getFluid(), aRecipe.mFluidOutputs[i], F) || mTanksOutput[i].getFluidAmount() > Math.max(999, aRecipe.mFluidOutputs[i].amount)))) {
               	return F;
        	}
    	}
    	return T;
	}
	
	/** return codes for checkRecipe() */
	protected static final int
	DID_NOT_FIND_RECIPE = 0,
	FOUND_RECIPE_BUT_DID_NOT_MEET_REQUIREMENTS = 1,
	FOUND_AND_SUCCESSFULLY_USED_RECIPE = 2,
	FOUND_AND_COULD_HAVE_USED_RECIPE = 3;
	
	/**
	 * Override this to check the Recipes yourself, super calls to this could be useful if you just want to add a special case
	 * I thought about Enum too, but Enum doesn't add support for people adding other return Systems.
	 * Funny how Eclipse marks the word Enum as not correctly spelled.
	 * @return see constants above
	 */
	protected int checkRecipe(boolean aApplyRecipe, boolean aUseAutoInputs) {
		mCouldUseRecipe = F;
    	if (mRecipes == null) return DID_NOT_FIND_RECIPE;
    	
    	DelegatorTileEntity<IInventory> tInventory = null;
    	byte tAutoInput = FACING_TO_SIDE[mFacing][mItemAutoInput];
		
    	if (aUseAutoInputs && SIDES_VALID[tAutoInput] && !mDisabledItemInput) {
    		tInventory = getAdjacentInventory(tAutoInput);
    		if (tInventory.mTileEntity == null) tInventory = null;
    	}
    	
		ItemStack[] tInputs = new ItemStack[mRecipes.mInputItemsCount];
		for (int i = 0; i < mRecipes.mInputItemsCount; i++) {
			if (tInventory != null) UT.Inventories.moveOneItemStackIntoSlot(tInventory.mTileEntity, this, tInventory.mSideOfTileEntity, i, null, F, (byte)64, (byte)1, (byte)64, (byte)1);
			tInputs[i] = slot(i);
		}
		
		FluidStack[] tInputFluids = new FluidStack[mTanksInput.length];
		for (int i = 0; i < mTanksInput.length; i++) {
			tInputFluids[i] = mTanksInput[i].getFluid();
		}
		
    	Recipe tRecipe = mRecipes.findRecipe(this, mLastRecipe, F, mEnergyTypeAccepted == TD.Energy.RF ? mInputMax / RF_PER_EU : mInputMax, slot(mRecipes.mInputItemsCount+mRecipes.mOutputItemsCount), tInputFluids, tInputs);
    	if (tRecipe == null) return DID_NOT_FIND_RECIPE;
    	if (tRecipe.mCanBeBuffered) mLastRecipe = tRecipe;
    	if (!canOutput(tRecipe)) return FOUND_RECIPE_BUT_DID_NOT_MEET_REQUIREMENTS;
    	if (!tRecipe.isRecipeInputEqual(aApplyRecipe, F, tInputFluids, tInputs)) return FOUND_RECIPE_BUT_DID_NOT_MEET_REQUIREMENTS;
    	mCouldUseRecipe = T;
    	if (!aApplyRecipe) return FOUND_AND_COULD_HAVE_USED_RECIPE;
    	
    	for (byte tSide : ALL_SIDES_VALID_FIRST[FACING_TO_SIDE[mFacing][mItemAutoInput]]) if (FACE_CONNECTED[FACING_ROTATIONS[mFacing][tSide]][mItemInputs]) {
    		DelegatorTileEntity<TileEntity> tDelegator = getAdjacentTileEntity(tSide);
    		if (tDelegator.mTileEntity instanceof ITileEntityAdjacentInventoryUpdatable) {
    			((ITileEntityAdjacentInventoryUpdatable)tDelegator.mTileEntity).adjacentInventoryUpdated(tDelegator.mSideOfTileEntity, this);
    		}
    	}
    	
        mOutputItems = tRecipe.getOutputs(worldObj.rand);
        mOutputFluids = tRecipe.getFluidOutputs(worldObj.rand);
        mMinEnergy = Math.max(1, (mEnergyTypeAccepted == TD.Energy.RF ? tRecipe.mEUt * RF_PER_EU : tRecipe.mEUt));
        mMaxProgress = Math.max(1, UT.Code.units(mMinEnergy * Math.max(1, tRecipe.mDuration), mEfficiency, 10000, T));
        if (mMinEnergy > 0) while (mMinEnergy < mInputMin && mMinEnergy * 4 <= mInputMax) {mMinEnergy *= 4; mMaxProgress *= 2;}
        
        removeAllDroppableNullStacks();
        for (int i = 0; i < mTanksInput.length; i++) if (mTanksInput[i].getFluidAmount() == 0) mTanksInput[i].setFluid(null);
        return FOUND_AND_SUCCESSFULLY_USED_RECIPE;
    }
	
	protected void doWork(long aTimer) {
		mSuccessful = F;
		if (mEnergy >= mInputMin && mEnergy >= mMinEnergy) {
			mActive = doActive(aTimer, Math.min(mInputMax, mEnergy));
			mRunning = T;
		} else {
			mActive = doInactive(aTimer);
			mRunning = F;
		}
		mEnergy -= mInputMax; if (mEnergy < 0) mEnergy = 0;
	}
	
	protected boolean doActive(long aTimer, long aEnergy) {
		boolean rActive = F;
		
		if (mMaxProgress > 0) {
			rActive = T;
			if (mProgress <= mMaxProgress) mProgress += aEnergy;
    		if (mProgress >= mMaxProgress && (mStateOld&&!mStateNew || !TD.Energy.ALL_ALTERNATING.contains(mEnergyTypeAccepted))) {
    			for (int i = 0; i < mRecipes.mOutputItemsCount; i++) for (int j = 0; j < mOutputItems .length; j++) if (mOutputItems [j] != null && addStackToSlot(mRecipes.mInputItemsCount+i, mOutputItems[j])) {mSuccessful = T; mOutputItems[j] = null; break;}
    			for (int i = 0; i < mTanksOutput.length; i++) for (int j = 0; j < mOutputFluids.length; j++) if (mOutputFluids[j] != null) {
    				if (mTanksOutput[i].getFluidAmount() == 0) {
    					mTanksOutput[i].setFluid(mOutputFluids[j]);
            			mSuccessful = T; mOutputFluids[j] = null; break;
    				} else if (UT.Fluids.equal(mTanksOutput[i].getFluid(), mOutputFluids[j], F)) {
    					updateInventory();
    					mTanksOutput[i].getFluid().amount += mOutputFluids[j].amount;
            			mSuccessful = T; mOutputFluids[j] = null; break;
    				}
    			}
    			
    			if (UT.Code.containsSomething(mOutputItems) || UT.Code.containsSomething(mOutputFluids)) {
        			mMinEnergy = 0;
        			mProgress = mMaxProgress;
    			} else {
        			mProgress -= mMaxProgress; // this way the leftover energy can be used on the next processed thing.
        			mMinEnergy = 0;
        			mMaxProgress = 0;
        			mOutputItems = ZL_ITEMSTACK;
        			mOutputFluids = ZL_FLUIDSTACK;
        			mSuccessful = T;
        			
        	    	for (byte tSide : ALL_SIDES_VALID_FIRST[FACING_TO_SIDE[mFacing][mItemAutoOutput]]) if (FACE_CONNECTED[FACING_ROTATIONS[mFacing][tSide]][mItemOutputs]) {
        	    		DelegatorTileEntity<TileEntity> tDelegator = getAdjacentTileEntity(tSide);
        	    		if (tDelegator.mTileEntity instanceof ITileEntityAdjacentInventoryUpdatable) {
        	    			((ITileEntityAdjacentInventoryUpdatable)tDelegator.mTileEntity).adjacentInventoryUpdated(tDelegator.mSideOfTileEntity, this);
        	    		}
        	    	}
        			
        			onProcessFinished();
    			}
    		}
    	}
		
		mStateOld = mStateNew;
		
		if (mSuccessful || aTimer%20 == 5) {
			if (!mDisabledFluidInput && SIDES_VALID[mFluidAutoInput]) {
				byte tAutoInput = FACING_TO_SIDE[mFacing][mFluidAutoInput];
				DelegatorTileEntity<IFluidHandler> tTileEntity = getAdjacentTank(tAutoInput);
				if (tTileEntity.mTileEntity != null) {
					FluidTankInfo[] tInfos = tTileEntity.mTileEntity.getTankInfo(FORGE_DIR[tTileEntity.mSideOfTileEntity]);
					if (tInfos != null) for (FluidTankInfo tInfo : tInfos) if (tInfo != null && tInfo.fluid != null && tInfo.fluid.amount > 0 && getFluidTankFillable(SIDE_ANY, tInfo.fluid) != null) {
						FluidStack tDrained = tTileEntity.mTileEntity.drain(FORGE_DIR[tTileEntity.mSideOfTileEntity], tInfo.fluid, F);
						if (tDrained != null) {
							tDrained.amount = fill(FORGE_DIR[tAutoInput], tDrained, F);
							if (tDrained.amount > 0) fill(FORGE_DIR[tAutoInput], tTileEntity.mTileEntity.drain(FORGE_DIR[tTileEntity.mSideOfTileEntity], tDrained, T), T);
						}
					}
				}
			}
		}
		
		if (!mDisabledItemOutput && SIDES_VALID[mItemAutoOutput]) {
			boolean
			tOutputEmpty = T;
			for (int i = mRecipes.mInputItemsCount, j = i + mRecipes.mOutputItemsCount; i < j; i++) if (slot(i) != null) {tOutputEmpty = F; break;}
			
	    	// Output not Empty && (Successfully produced something || Some Inventory Stuff changes || The Machine has just been turned ON || Output has been blocked since 256 active ticks || Check ones every Minute)
			if (!tOutputEmpty && (mSuccessful || mInventoryChanged || !mRunning || mOutputBlocked == 1 || aTimer%1200 == 5)) {
				byte tAutoOutput = FACING_TO_SIDE[mFacing][mItemAutoOutput];
				DelegatorTileEntity<TileEntity> tTileEntity = getAdjacentTileEntity(tAutoOutput);
				for (int i = 0; i < mRecipes.mOutputItemsCount; i++) {
					UT.Inventories.moveOneItemStack(this, tTileEntity, tAutoOutput, tTileEntity.mSideOfTileEntity, null, F, 64, 1, 64, 1);
				}
			}
			
			tOutputEmpty = T;
			for (int i = mRecipes.mInputItemsCount, j = i + mRecipes.mOutputItemsCount; i < j; i++) if (slot(i) != null) {tOutputEmpty = F; mOutputBlocked++; break;}
			
			if (tOutputEmpty) mOutputBlocked = 0;
		}
		
    	if (mMaxProgress <= 0) {
    		// Successfully produced something || Some Inventory Stuff changes || The Machine has just been turned ON || Check ones every Minute
    		if ((mSuccessful || mInventoryChanged || !mRunning || aTimer%1200 == 5) && checkRecipe(T, T) == FOUND_AND_SUCCESSFULLY_USED_RECIPE) {
		        onProcessStarted();
	    	} else {
	    		mProgress = 0;
	    	}
		}
		
		return rActive;
	}
	
	protected boolean doInactive(long aTimer) {
		if (mProgress > 0 || mActive) UT.Sounds.send(worldObj, SFX.IC_MACHINE_INTERRUPT, 1.0F, 1.0F, xCoord, yCoord, zCoord);
		mProgress = 0;
		if (mRunning || mInventoryChanged || aTimer%1200 == 5) checkRecipe(F, T);
		return F;
	}
	
	public class MultiTileEntityGUICommonBasicMachine extends ContainerCommon {
		@SuppressWarnings("hiding")
		private RecipeMap mRecipes;
		
		public MultiTileEntityGUICommonBasicMachine(InventoryPlayer aInventoryPlayer, MultiTileEntityBasicMachine aTileEntity) {
			super(aInventoryPlayer, aTileEntity);
			mRecipes = aTileEntity.mRecipes;
		}
		
	    @Override
		public int addSlots(InventoryPlayer aPlayerInventory) {
	        int tIndex = 0;
	        
			mRecipes = ((MultiTileEntityBasicMachine)mTileEntity).mRecipes;
	        
	        switch (mRecipes.mInputItemsCount) {
	        case  0:
	        	break;
	        case  1:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>6? 7:25));
	        	break;
	        case  2:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, mRecipes.mInputFluidCount>6? 7:25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>6? 7:25));
	        	break;
	        case  3:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, mRecipes.mInputFluidCount>6? 7:25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, mRecipes.mInputFluidCount>6? 7:25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>6? 7:25));
	        	break;
	        case  4:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, mRecipes.mInputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, mRecipes.mInputFluidCount>3?25:34));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>3?25:34));
	        	break;
	        case  5:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, mRecipes.mInputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, mRecipes.mInputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, mRecipes.mInputFluidCount>3?25:34));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>3?25:34));
	        	break;
	        case  6:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, mRecipes.mInputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, mRecipes.mInputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, mRecipes.mInputFluidCount>3?25:34));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, mRecipes.mInputFluidCount>3?25:34));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, mRecipes.mInputFluidCount>3?25:34));
	        	break;
	        case  7:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 43));
	        	break;
	        case  8:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 43));
	        	break;
	        case  9:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 43));
	        	break;
	        case 10:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 61));
	        	break;
	        case 11:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 61));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 61));
	        	break;
	        default:
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 43));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 17, 61));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 35, 61));
	            addSlotToContainer(new Slot_Normal(mTileEntity, tIndex++, 53, 61));
	        	break;
	        }
	        
	        switch (mRecipes.mOutputItemsCount) {
	        case  0:
	        	break;
	        case  1:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>6? 7:25));
	        	break;
	        case  2:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>6? 7:25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, mRecipes.mOutputFluidCount>6? 7:25));
	        	break;
	        case  3:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>6? 7:25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, mRecipes.mOutputFluidCount>6? 7:25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, mRecipes.mOutputFluidCount>6? 7:25));
	        	break;
	        case  4:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, mRecipes.mOutputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>3?25:34));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, mRecipes.mOutputFluidCount>3?25:34));
	        	break;
	        case  5:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, mRecipes.mOutputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, mRecipes.mOutputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>3?25:34));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, mRecipes.mOutputFluidCount>3?25:34));
	        	break;
	        case  6:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, mRecipes.mOutputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, mRecipes.mOutputFluidCount>3? 7:16));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, mRecipes.mOutputFluidCount>3?25:34));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, mRecipes.mOutputFluidCount>3?25:34));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, mRecipes.mOutputFluidCount>3?25:34));
	        	break;
	        case  7:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	        	break;
	        case  8:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	        	break;
	        case  9:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 43));
	        	break;
	        case 10:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 61));
	        	break;
	        case 11:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 61));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 61));
	        	break;
	        default:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 61));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 61));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 61));
	        	break;
	        }
	        
	        tIndex++; // Special Slot.
	        
	        for (int i = 0; i < mRecipes.mInputFluidCount ; i++) addSlotToContainer(new Slot_Render(mTileEntity, tIndex++,  53 - (i%3)*18, 63 - (i/3)*18));
	        for (int i = 0; i < mRecipes.mOutputFluidCount; i++) addSlotToContainer(new Slot_Render(mTileEntity, tIndex++, 107 + (i%3)*18, 63 - (i/3)*18));
	        
	        return super.addSlots(aPlayerInventory);
	    }
	    
	    public short mProgressBar = 0;
	    
	    @Override
		public void detectAndSendChanges() {
	        super.detectAndSendChanges();
	    	for (ICrafting tUpdate : (List<ICrafting>)crafters) {
	    		if (((MultiTileEntityBasicMachine)mTileEntity).mSuccessful) {
	    			tUpdate.sendProgressBarUpdate(this, 0, Short.MAX_VALUE);
	    		} else if (((MultiTileEntityBasicMachine)mTileEntity).mMaxProgress > 0) {
	    			tUpdate.sendProgressBarUpdate(this, 0, (short)UT.Code.units(Math.min(((MultiTileEntityBasicMachine)mTileEntity).mMaxProgress, ((MultiTileEntityBasicMachine)mTileEntity).mProgress), ((MultiTileEntityBasicMachine)mTileEntity).mMaxProgress, Short.MAX_VALUE, T));
	    		} else {
	    			tUpdate.sendProgressBarUpdate(this, 0, -1);
	    		}
	    	}
	    }
	    
	    @Override
		@SideOnly(Side.CLIENT)
	    public void updateProgressBar(int aIndex, int aValue) {
	    	super.updateProgressBar(aIndex, aValue);
	    	switch (aIndex) {
	    	case 0: mProgressBar = (short)aValue; break;
	    	}
	    }
	    
	    @Override public int getSlotStartIndex() {return 0;}
	    @Override public int getShiftClickStartIndex() {return 0;}
	    @Override public int getSlotCount() {return mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount;}
	    @Override public int getShiftClickSlotCount() {return mRecipes.mInputItemsCount;}
	}
	
	@SideOnly(Side.CLIENT)
	public class MultiTileEntityGUIClientBasicMachine extends ContainerClient {
		public final MultiTileEntityBasicMachine mTileEntity;
		
	    public MultiTileEntityGUIClientBasicMachine(InventoryPlayer aInventoryPlayer, MultiTileEntityBasicMachine aTileEntity) {
	        super(new MultiTileEntityGUICommonBasicMachine(aInventoryPlayer, aTileEntity), aTileEntity.mGUITexture);
	        mTileEntity = aTileEntity;
	        mNEI = mTileEntity.mRecipes.mNameNEI;
	    }
	    
	    @Override
	    protected void drawGuiContainerForegroundLayer(int par1, int par2) {
	    	fontRendererObj.drawString(mTileEntity.hasCustomInventoryName()?mTileEntity.getInventoryNameGUI():LH.get(mTileEntity.mRecipes.mNameInternal), 8,  4, 4210752);
	    }
	    
	    @Override
	    protected void drawGuiContainerBackgroundLayer2(float par1, int par2, int par3) {
	        int x = (width - xSize) / 2;
	        int y = (height - ySize) / 2;
	        drawTexturedModalRect(x, y, 0, 0, xSize, ySize);
	        if (mContainer instanceof MultiTileEntityGUICommonBasicMachine) {
	        	if (((MultiTileEntityGUICommonBasicMachine)mContainer).mProgressBar >= 0) {
		        	int tSize = (mTileEntity.mRecipes.mProgressBarDirection % 4 < 2 ? 20 : 18), tProgress = (int)UT.Code.scale(((MultiTileEntityGUICommonBasicMachine)mContainer).mProgressBar, 0, Short.MAX_VALUE, tSize * UT.Code.unsignB(mTileEntity.mRecipes.mProgressBarAmount), F) % (tSize+1);
		        	
		        	switch (mTileEntity.mRecipes.mProgressBarDirection) {
		        	case 0:								drawTexturedModalRect(x + 78					, y + 24					, 176					, 0					, tProgress	, 18		); break;
		        	case 1:								drawTexturedModalRect(x + 78 + 20 - tProgress	, y + 24					, 176 + 20 - tProgress	, 0					, tProgress	, 18		); break;
		        	case 2:								drawTexturedModalRect(x + 78					, y + 24					, 176					, 0					, 20		, tProgress	); break;
		        	case 3:								drawTexturedModalRect(x + 78					, y + 24 + 18 - tProgress	, 176					, 18 - tProgress	, 20		, tProgress	); break;
		        	case 4: tProgress = 20 - tProgress;	drawTexturedModalRect(x + 78					, y + 24					, 176					, 0					, tProgress	, 18		); break;
		        	case 5: tProgress = 20 - tProgress;	drawTexturedModalRect(x + 78 + 20 - tProgress	, y + 24					, 176 + 20 - tProgress	, 0					, tProgress	, 18		); break;
		        	case 6: tProgress = 18 - tProgress;	drawTexturedModalRect(x + 78					, y + 24					, 176					, 0					, 20		, tProgress	); break;
		        	case 7: tProgress = 18 - tProgress;	drawTexturedModalRect(x + 78					, y + 24 + 18 - tProgress	, 176					, 18 - tProgress	, 20		, tProgress	); break;
		        	}
	        	}
	        }
	    }
	}
	
	protected void onProcessStarted() {/**/}
	protected void onProcessFinished() {/**/}
	
	@Override public Object getGUIClient2(int aGUIID, EntityPlayer aPlayer) {return new MultiTileEntityGUIClientBasicMachine(aPlayer.inventory, this);}
	@Override public Object getGUIServer2(int aGUIID, EntityPlayer aPlayer) {return new MultiTileEntityGUICommonBasicMachine(aPlayer.inventory, this);}
	
	@Override public byte getVisualData() {return (byte)((mActive?1:0)|(mRunning?2:0));}
	
	@Override public void setVisualData(byte aData) {mRunning=((aData&2)!=0); mActive=((aData&1)!=0);}
	@Override public byte getDefaultSide() {return SIDE_SOUTH;}
	@Override public boolean[] getValidSides() {return SIDES_HORIZONTAL;}
	@Override public ITexture getTexture2(Block aBlock, int aRenderPass, byte aSide, boolean[] aShouldSideBeRendered) {return aShouldSideBeRendered[aSide] ? BlockTextureMulti.get(BlockTextureDefault.get(mTexturesMaterial[FACING_ROTATIONS[mFacing][aSide]], mRGBa), BlockTextureDefault.get((mActive||worldObj==null?mTexturesActive:mRunning?mTexturesRunning:mTexturesInactive)[FACING_ROTATIONS[mFacing][aSide]])) : null;}
	
	@Override public boolean hasWork() {return mMaxProgress > 0;}
	@Override public long getProgressValue(byte aSide) {return mSuccessful ? getProgressMax(aSide) : mMinEnergy < 1 ? mProgress    : mProgress    / mMinEnergy + (mProgress    % mMinEnergy == 0 ? 0 : 1) ;}
	@Override public long getProgressMax  (byte aSide) {return Math.max(1,                           mMinEnergy < 1 ? mMaxProgress : mMaxProgress / mMinEnergy + (mMaxProgress % mMinEnergy == 0 ? 0 : 1));}
	
	@Override public boolean getStateRunningPossible() {return mCouldUseRecipe || mActive;}
	@Override public boolean getStateRunningPassively() {return mRunning;}
	@Override public boolean getStateRunningActively() {return mActive;}
	@Override public boolean setStateOnOff(boolean aOnOff) {if (mStopped == aOnOff) {mStopped = !aOnOff; updateAdjacentToggleableEnergySources();} else mStopped = !aOnOff; return !mStopped;}
	@Override public boolean getStateOnOff() {return !mStopped;}
	
	@Override public String getTileEntityName() {return "gt.multitileentity.machine.basic";}
	
	@Override
	public void adjacentInventoryUpdated(byte aSide, IInventory aTileEntity) {
		if (FACE_CONNECTED[FACING_ROTATIONS[mFacing][aSide]][mItemInputs|mItemOutputs]) updateInventory();
	}
}